@@ -1,19 +1,29 @@
 /*
  * Copyright 2003-2006 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.129 2006/02/18 15:51:11 solar Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.137 2006/03/19 02:41:17 solar Exp $
  *
  * Copyright 2003-2006 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2004-2006 Mike Frysinger  - <vapier@gentoo.org>
  */
 
 #include "paxinc.h"
-
-static const char *rcsid = "$Id: scanelf.c,v 1.129 2006/02/18 15:51:11 solar Exp $";
+#ifdef __linux__
+ #include <glob.h>
+#endif
+static const char *rcsid = "$Id: scanelf.c,v 1.137 2006/03/19 02:41:17 solar Exp $";
 #define argv0 "scanelf"
 
-#define IS_MODIFIER(c) (c == '%' || c == '#')
+#define IS_MODIFIER(c) (c == '%' || c == '#' || c == '+')
 
+#define do_state(option, flag)		\
+	if (islower(option)) {		\
+		flags &= ~PF_##flag;	\
+		flags |= PF_NO##flag;	\
+	} else {			\
+		flags &= ~PF_NO##flag;	\
+		flags |= PF_##flag;	\
+	}
 
 
 /* prototypes */
@@ -54,6 +64,7 @@ static char show_banner = 1;
 static char be_quiet = 0;
 static char be_verbose = 0;
 static char be_wewy_wewy_quiet = 0;
+static char be_semi_verbose = 0;
 static char *find_sym = NULL, *versioned_symname = NULL;
 static char *find_lib = NULL;
 static char *find_section = NULL;
@@ -204,7 +215,7 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 			flags = EGET(phdr[i].p_flags); \
 			if (be_quiet && ((flags & check_flags) != check_flags)) \
 				continue; \
-			if (fix_elf && ((flags & PF_X) != flags)) { \
+			if ((EGET(phdr[i].p_type) != PT_LOAD) && (fix_elf && ((flags & PF_X) != flags))) { \
 				ESET(phdr[i].p_flags, flags & (PF_X ^ (size_t)-1)); \
 				ret[3] = ret[7] = '!'; \
 				flags = EGET(phdr[i].p_flags); \
@@ -735,6 +746,7 @@ static char *scanelf_file_bind(elfobj *elf, char *found_bind)
 {
 	unsigned long i;
 	struct stat s;
+	char dynamic = 0;
 
 	if (!show_bind) return NULL;
 	if (!elf->phdr) return NULL;
@@ -747,6 +759,7 @@ static char *scanelf_file_bind(elfobj *elf, char *found_bind)
 		Elf ## B ## _Off offset; \
 		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
 			if (EGET(phdr[i].p_type) != PT_DYNAMIC) continue; \
+			dynamic = 1; \
 			offset = EGET(phdr[i].p_offset); \
 			if (offset >= elf->len - sizeof(Elf ## B ## _Dyn)) continue; \
 			dyn = DYN ## B (elf->data + offset); \
@@ -771,7 +784,7 @@ static char *scanelf_file_bind(elfobj *elf, char *found_bind)
 		return NULL;
 	} else {
 		*found_bind = 1;
-		return (char *) "LAZY";
+		return (char *) (dynamic ? "LAZY" : "STATIC");
 	}
 }
 static char *scanelf_file_soname(elfobj *elf, char *found_soname)
@@ -843,12 +856,15 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 			cnt = EGET(symtab->sh_size) / cnt; \
 		for (i = 0; i < cnt; ++i) { \
 			if (sym->st_name) { \
+				/* make sure the symbol name is in acceptable memory range */ \
 				symname = (char *)(elf->data + EGET(strtab->sh_offset) + EGET(sym->st_name)); \
 				if ((void*)symname > (void*)elf->data_end) { \
 					warnf("%s: corrupt ELF symbols", elf->filename); \
+					++sym; \
 					continue; \
 				} \
-				if (*find_sym == '*') { \
+				/* debug display ... show all symbols and some extra info */ \
+				if (*ret == '*') { \
 					printf("%s(%s) %5lX %15s %s\n", \
 					       ((*found_sym == 0) ? "\n\t" : "\t"), \
 					       elf->base_filename, \
@@ -857,19 +873,38 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 					       symname); \
 					*found_sym = 1; \
 				} else { \
+					/* allow the user to specify a comma delimited list of symbols to search for */ \
 					char *this_sym, *next_sym; \
-					this_sym = find_sym; \
+					this_sym = ret; \
 					do { \
 						next_sym = strchr(this_sym, ','); \
 						if (next_sym == NULL) \
 							next_sym = this_sym + strlen(this_sym); \
+						/* do we want a defined symbol ? */ \
+						if (*this_sym == '+') { \
+							if (sym->st_value == 0) \
+								goto skip_this_sym##B; \
+							++this_sym; \
+						/* do we want an undefined symbol ? */ \
+						} else if (*this_sym == '-') { \
+							if (sym->st_value != 0) \
+								goto skip_this_sym##B; \
+							++this_sym; \
+						} \
+						/* ok, lets compare the name now */ \
 						if ((strncmp(this_sym, symname, (next_sym-this_sym)) == 0 && symname[next_sym-this_sym] == '\0') || \
 						    (strcmp(symname, versioned_symname) == 0)) { \
-							ret = this_sym; \
+							if (be_semi_verbose) { \
+								char buf[126]; \
+								snprintf(buf, sizeof(buf), "%lX %s %s", \
+									(unsigned long)sym->st_size, get_elfstttype(sym->st_info), this_sym); \
+								ret = buf; \
+							} else \
+								ret = this_sym; \
 							(*found_sym)++; \
 							goto break_out; \
 						} \
-						this_sym = next_sym + 1; \
+						skip_this_sym##B: this_sym = next_sym + 1; \
 					} while (*next_sym != '\0'); \
 				} \
 			} \
@@ -898,9 +933,11 @@ static char *scanelf_file_sections(elfobj *elf, char *found_section)
 
 #define FIND_SECTION(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
+		int invert; \
 		Elf ## B ## _Shdr *section; \
-		section = SHDR ## B (elf_findsecbyname(elf, find_section)); \
-		if (section != NULL) \
+		invert = (*find_section == '!' ? 1 : 0); \
+		section = SHDR ## B (elf_findsecbyname(elf, find_section+invert)); \
+		if ((section == NULL && invert) || (section != NULL && !invert)) \
 			*found_section = 1; \
 	}
 	FIND_SECTION(32)
@@ -953,6 +990,7 @@ static int scanelf_elfobj(elfobj *elf)
 			if (!IS_MODIFIER(out_format[i])) continue;
 
 			switch (out_format[++i]) {
+			case '+': break;
 			case '%': break;
 			case '#': break;
 			case 'F':
@@ -992,7 +1030,9 @@ static int scanelf_elfobj(elfobj *elf)
 
 		out = NULL;
 		be_wewy_wewy_quiet = (out_format[i] == '#');
+		be_semi_verbose = (out_format[i] == '+');
 		switch (out_format[++i]) {
+		case '+':
 		case '%':
 		case '#':
 			xchrcat(&out_buffer, out_format[i], &out_len); break;
@@ -1215,13 +1255,13 @@ static void scanelf_dir(const char *path)
 	closedir(dir);
 }
 
-static int scanelf_from_file(char *filename)
+static int scanelf_from_file(const char *filename)
 {
 	FILE *fp = NULL;
 	char *p;
 	char path[__PAX_UTILS_PATH_MAX];
 
-	if (((strcmp(filename, "-")) == 0) && (ttyname(0) == NULL))
+	if (strcmp(filename, "-") == 0)
 		fp = stdin;
 	else if ((fp = fopen(filename, "r")) == NULL)
 		return 1;
@@ -1254,7 +1294,7 @@ static int load_ld_so_conf(int i, const char *fname)
 			*p = 0;
 		if ((p = strchr(path, '\n')) != NULL)
 			*p = 0;
-#ifdef HAVE_GLOB
+#ifdef __linux__
 		// recursive includes of the same file will make this segfault.
 		if ((memcmp(path, "include", 7) == 0) && isblank(path[7])) {
 			glob64_t gl;
@@ -1456,7 +1496,7 @@ static void usage(int status)
 static void parseargs(int argc, char *argv[])
 {
 	int i;
-	char *from_file = NULL;
+	const char *from_file = NULL;
 
 	opterr = 0;
 	while ((i=getopt_long(argc, argv, PARSE_FLAGS, long_opts, NULL)) != -1) {
@@ -1512,15 +1552,6 @@ static void parseargs(int argc, char *argv[])
 			unsigned long flags = (PF_NOEMUTRAMP | PF_NORANDEXEC);
 			size_t x;
 
-#define do_state(option, flag)				\
-			if (islower(option)) {		\
-				flags &= ~PF_##flag;	\
-				flags |= PF_NO##flag;	\
-			} else {			\
-				flags &= ~PF_NO##flag;	\
-				flags |= PF_##flag;	\
-			}
-
 			for (x = 0 ; x < strlen(optarg); x++) {
 				switch(optarg[x]) {
 					case 'p':
@@ -1601,6 +1632,7 @@ static void parseargs(int argc, char *argv[])
 			if (!IS_MODIFIER(out_format[i])) continue;
 
 			switch (out_format[++i]) {
+			case '+': break;
 			case '%': break;
 			case '#': break;
 			case 'F': break;
@@ -1651,6 +1683,8 @@ static void parseargs(int argc, char *argv[])
 		load_ld_so_conf(0, "/etc/ld.so.conf");
 	if (scan_ldpath) scanelf_ldpath();
 	if (scan_envpath) scanelf_envpath();
+	if (!from_file && ttyname(0) == NULL)
+		from_file = "-";
 	if (from_file) {
 		scanelf_from_file(from_file);
 		from_file = *argv;
